--------------------------------------------------------------------------------
-- From Structured to Deriving0
--------------------------------------------------------------------------------

{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE MultiParamTypeClasses  #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE UndecidableInstances   #-}


module Conversion.S2D0 where

import qualified Library.Structured as S
import qualified Library.Deriving   as D
-- import qualified Conversion.S2S0    as C

--------------------------------------------------------------------------------
-- Conversion of representations (types)
--------------------------------------------------------------------------------

type family S2D0 (r :: k) (p :: *) :: D.Un * -- overly polymorphic, for overloading
type instance S2D0 ('S.UData md cs) p = S2D0 cs p

-- Constructors
type instance S2D0 (S.Empty               :: S.Tree (S.UCon k)) p = D.VV
type instance S2D0 (S.Leaf ('S.UCon mc x) :: S.Tree (S.UCon k)) p = D.MM mc (S2D0 x p)
type instance S2D0 (S.Bin x y             :: S.Tree (S.UCon k)) p = (D.:+:) (S2D0 x p) (S2D0 y p)

-- Arguments to constructors
type instance S2D0 (S.Empty               :: S.Tree (S.USel k)) p = D.UU
type instance S2D0 (S.Leaf ('S.USel ms x) :: S.Tree (S.USel k)) p = S2D0 x p
type instance S2D0 (S.Bin x y             :: S.Tree (S.USel k)) p = (D.:**:) (S2D0 x p) (S2D0 y p)

-- Fields
type instance S2D0 (S.UK1 i a)    p = D.KK i       a
type instance S2D0 (S.URec i f)   p = D.KK (D.R i) (f p)
type instance S2D0 (S.UPar)       p = D.KK D.P     p
type instance S2D0 ((S.:.:) f a)  p = D.KK D.U     (f (S2D0_Comp a p))

type family S2D0_Comp (r :: S.UArg *) (p :: *) :: *
type instance S2D0_Comp S.UPar         p = p
type instance S2D0_Comp (S.UK1 i a)    p = a
type instance S2D0_Comp (S.URec i f)   p = f p
type instance S2D0_Comp ((S.:.:) f r)  p = f (S2D0_Comp r p)

--------------------------------------------------------------------------------
-- Value-level conversion
--------------------------------------------------------------------------------

class Convert (r :: k) where
  s2d0 :: S.In r p -> D.In (S2D0 r p) p
  d02s :: D.In (S2D0 r p) p -> S.In r p

instance (Convert cs) => Convert ('S.UData md cs) where
  s2d0 (S.D1 x) = s2d0 x
  d02s x = S.D1 (d02s x)

-- Constructors: none
instance Convert (S.Empty :: S.Tree (S.UCon k)) where
  s2d0 _ = error "no element of type void"
  d02s _ = error "no element of type void"

-- Constructors: picking one
instance (Convert x, Convert y)
        => Convert (S.Bin x y :: S.Tree (S.UCon k)) where
  s2d0 (S.L1 x) = D.L1 (s2d0 x)
  s2d0 (S.R1 x) = D.R1 (s2d0 x)
  d02s (D.L1 x) = S.L1 (d02s x)
  d02s (D.R1 x) = S.R1 (d02s x)

-- Constructors: metadata, proceed to constructor arguments
instance (Convert x)
        => Convert (S.Leaf ('S.UCon mc x) :: S.Tree (S.UCon k)) where
  s2d0 (S.C1 x) = D.M1 (s2d0 x)
  d02s (D.M1 x) = S.C1 (d02s x)

-- Constructor arguments: none
instance Convert (S.Empty :: S.Tree (S.USel k)) where
  s2d0 S.U1 = D.U1
  d02s D.U1 = S.U1

-- Constructor arguments: tupling
instance (Convert l, Convert r)
        => Convert (S.Bin l r :: S.Tree (S.USel k)) where
  s2d0 ((S.:*:) x y) = (D.:*:) (s2d0 x) (s2d0 y)
  d02s ((D.:*:) x y) = (S.:*:) (d02s x) (d02s y)

-- Constructor arguments: metadata, proceed to actual argument
instance (Convert x)
        => Convert (S.Leaf ('S.USel ms x) :: S.Tree (S.USel k)) where
  s2d0 (S.S1 x) = s2d0 x
  d02s x        = S.S1 (d02s x)

-- Actual arguments
instance Convert (S.UK1 i a :: S.UArg *) where
  s2d0 (S.K1 x) = D.K1 x
  d02s (D.K1 x) = S.K1 x

instance Convert (S.URec i a :: S.UArg *) where
  s2d0 (S.Rec x) = D.K1 x
  d02s (D.K1 x) = S.Rec x

instance Convert (S.UPar :: S.UArg *) where
  s2d0 (S.Par p) = D.K1 p
  d02s (D.K1 p) = S.Par p

instance (Functor t, Convert_Comp f)
        => Convert ((S.:.:) t f :: S.UArg *) where
  s2d0 (S.Comp p) = D.K1 (s2s0_Comp p)
  d02s (D.K1 p) = S.Comp (s02s_Comp p)

class Convert_Comp (r :: S.UArg *) where
  s2s0_Comp :: Functor f => f (S.In r p) -> f (S2D0_Comp r p)
  s02s_Comp :: Functor f => f (S2D0_Comp r p) -> f (S.In r p)

instance Convert_Comp S.UPar where
  s2s0_Comp = fmap S.unPar
  s02s_Comp = fmap S.Par

instance Convert_Comp (S.UK1 i a) where
  s2s0_Comp = fmap S.unK1
  s02s_Comp = fmap S.K1

instance Convert_Comp (S.URec i f) where
  s2s0_Comp = fmap S.unRec
  s02s_Comp = fmap S.Rec

instance (Functor f, Convert_Comp r) => Convert_Comp ((S.:.:) f r) where
  s2s0_Comp = fmap (s2s0_Comp . S.unComp)
  s02s_Comp = fmap (S.Comp . s02s_Comp)

--------------------------------------------------------------------------------
-- Embedding conversion
--------------------------------------------------------------------------------

instance (S.Generic a, Convert (S.Rep a)) => D.Generic a where
  type Rep a = S2D0 (S.Rep a) (S.Par a)
  type Par a = S.Par a

  from x = s2d0 (S.from x)
  to   x = S.to (d02s x)
