--------------------------------------------------------------------------------
-- Testing Library.Deriving
--------------------------------------------------------------------------------

{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE ConstraintKinds        #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE MultiParamTypeClasses  #-}
{-# LANGUAGE OverlappingInstances   #-}
{-# LANGUAGE UndecidableInstances   #-}
{-# LANGUAGE DeriveDataTypeable     #-}

module Test.Deriving where


import Data.Typeable
import Library.Deriving


--------------------------------------------------------------------------------
-- Example encoding: Int
--------------------------------------------------------------------------------

instance Generic Int where
  type Rep Int = KK U Int
  from n = K1 n
  to (K1 n) = n

--------------------------------------------------------------------------------
-- Example encoding: lists
--------------------------------------------------------------------------------

instance Generic [a] where
  type Rep [a] =  MM (MC "[]" Prefix False) UU
              :+: MM (MC ":"  (Infix RightAssociative 5) False) 
                     (KK P a :**: KK (R S) [a])

  from []    = L1 (M1 U1)
  from (h:t) = R1 (M1 (K1 h :*: K1 t))

  to (L1 (M1 U1))              = []
  to (R1 (M1 (K1 h :*: K1 t))) = h:t


instance Generic1 [] where
  type Rep1 [] = UU :+: (PAR :**: REC S [])

  from1 []    = L1 U1
  from1 (h:t) = R1 (Par1 h :*: Rec1 t)

  to1 (L1 U1)                  = []
  to1 (R1 (Par1 h :*: Rec1 t)) = h:t

--------------------------------------------------------------------------------
-- Example encoding: rose trees
--------------------------------------------------------------------------------

data RTree a = RTree a [RTree a] deriving (Show, Typeable)

instance Generic (RTree a) where
  type Rep (RTree a) = KK P a :**: KK U [RTree a]

  from (RTree x xs)     = K1 x :*: K1 xs
  to   (K1 x :*: K1 xs) = RTree x xs


instance Generic1 RTree where
  type Rep1 RTree = PAR :**: ([] :.: REC S RTree)

  from1 (RTree x xs)          = Par1 x :*: Comp1 (fmap Rec1 xs)
  to1   (Par1 x :*: Comp1 xs) = RTree x (fmap unRec1 xs)

--------------------------------------------------------------------------------
-- Example encoding: an enumeration datatype
--------------------------------------------------------------------------------

data Choice = C1 | C2 | C3 | C4 deriving Typeable

instance Generic Choice where
  type Rep Choice = (UU :+: UU) :+: (UU :+: UU)
  from C1 = L1 (L1 U1)
  from C2 = L1 (R1 U1)
  from C3 = R1 (L1 U1)
  from C4 = R1 (R1 U1)
  to (L1 (L1 U1)) = C1
  to (L1 (R1 U1)) = C2
  to (R1 (L1 U1)) = C3
  to (R1 (R1 U1)) = C4

--------------------------------------------------------------------------------
-- Example encoding: a mutually-recursive family
--------------------------------------------------------------------------------

data Zig = Zig Zag | ZigEnd
data Zag = Zag Zig

instance Generic Zig where
  type Rep Zig = KK (R O) Zag :+: UU

  from (Zig zag) = L1 (K1 zag)
  from ZigEnd    = R1 U1
  to (L1 (K1 zag)) = Zig zag
  to (R1 U1)       = ZigEnd

instance Generic Zag where
  type Rep Zag = KK (R O) Zig

  from (Zag zig) = K1 zig
  to (K1 zig)  = Zag zig

zigZag :: Zig
zigZag = Zig (Zag (Zig (Zag ZigEnd)))


--------------------------------------------------------------------------------
-- Generic function instances and tests
--------------------------------------------------------------------------------

instance (GEq a) => GEq [a]
instance (GEq a) => GEq (RTree a)

instance GFunctor []
instance GFunctor RTree

-- Tests
tgeq1 = geq [1] [2::Int]
tgeq2 = geq [1] [1::Int]
tgeq3 = geq (RTree (0::Int) [RTree 1 []]) (RTree 0 [RTree 1 [], RTree 2 []])


-- Tests
tgmap1 = gmap (+1) [0..9]
tgmap2 = gmap (+1) (RTree 0 [RTree 1 [], RTree 2 []])
