--------------------------------------------------------------------------------
-- Test encodings in Structured
--------------------------------------------------------------------------------

{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE DefaultSignatures      #-}

module Test.Structured where


import Library.Structured
import Conversion.S2RS
-- import Conversion.S2S0

--------------------------------------------------------------------------------
-- Example encoding: lists
--------------------------------------------------------------------------------

instance Generic [a] where
  type Rep [a] =
        'UData (MD "[]" "Prelude")
           (Bin (Leaf ('UCon (MC "[]" Prefix False)
                             Empty))
                (Leaf ('UCon (MC ":" (Infix RightAssociative 5) False) 
                             (Bin (Leaf ('USel (MS Nothing) UPar))
                                  (Leaf ('USel (MS Nothing) (URec S [])))))))
  type Par [a] = a

  from []    = D1 (L1 (C1 U1))
  from (h:t) = D1 (R1 (C1 (S1 (Par h) :*: S1 (Rec t))))

  to (D1 (L1 (C1 U1)))                          = []
  to (D1 (R1 (C1 (S1 (Par h) :*: S1 (Rec t))))) = h:t

--------------------------------------------------------------------------------
-- Example encoding: choice
--------------------------------------------------------------------------------

data Choice = A | B | C | D

instance Generic Choice where
  type Rep Choice =
          'UData (MD "Choice" "StructuredTest")
             (Bin (Bin (Leaf ('UCon (MC "A" Prefix False) Empty))
                       (Leaf ('UCon (MC "B" Prefix False) Empty)))
                  (Bin (Leaf ('UCon (MC "C" Prefix False) Empty))
                       (Leaf ('UCon (MC "D" Prefix False) Empty))))

  from A = D1 (L1 (L1 (C1 U1)))
  from B = D1 (L1 (R1 (C1 U1)))
  from C = D1 (R1 (L1 (C1 U1)))
  from D = D1 (R1 (R1 (C1 U1)))

  to (D1 (L1 (L1 (C1 U1)))) = A
  to (D1 (L1 (R1 (C1 U1)))) = B
  to (D1 (R1 (L1 (C1 U1)))) = C
  to (D1 (R1 (R1 (C1 U1)))) = D

--------------------------------------------------------------------------------
-- Example encoding: something contrived
--------------------------------------------------------------------------------

data D f a b = D1_ Int (f a) | D2_ [D f a b] b

instance Generic (D f a b) where
  type Rep (D f a b) =
          'UData (MD "D" "StructuredTest")
             (Bin (Leaf ('UCon (MC "D1_" Prefix False)
                        (Bin (Leaf ('USel (MS Nothing) (UK1 (R O) Int)))
                             (Leaf ('USel (MS Nothing) (UK1 U (f a)))))))
                  (Leaf ('UCon (MC "D2_" Prefix False)
                        (Bin (Leaf ('USel (MS Nothing) ([] :.: URec S (D f a))))
                             (Leaf ('USel (MS Nothing) UPar))))))

  type Par (D f a b) = b

  from (D1_ i f) = D1 (L1 (C1 (S1 (K1 i)              :*: S1 (K1 f))))
  from (D2_ d b) = D1 (R1 (C1 (S1 (Comp (fmap Rec d)) :*: S1 (Par b))))

  to (D1 (L1 (C1 (S1 (K1 i)   :*: S1 (K1 f)))))  = D1_ i f
  to (D1 (R1 (C1 (S1 (Comp d) :*: S1 (Par b))))) = D2_ (fmap unRec d) b

--------------------------------------------------------------------------------
-- Testing length
--------------------------------------------------------------------------------
{-
instance GLength [a]
instance GLength Choice

tLength1, tLength2 :: Int
tLength1 = glength [A,B,C,D]
tLength2 = glength D
-}